#include "postgres.h"
#include "access/attnum.h"
#include "utils/relcache.h"
#include "access/reloptions.h"
#include "access/table.h"
#include "access/relation.h"
#include "access/relscan.h"
#include "access/xact.h"
#include "commands/defrem.h"
#include "funcapi.h"
#include "utils/rel.h"
#include "access/nbtree.h"
#include "commands/vacuum.h"
#include "nodes/makefuncs.h"
#include "catalog/dependency.h"
#include "catalog/pg_operator.h"
#include "catalog/index.h"
#include "catalog/namespace.h"
#include "catalog/storage.h"
#include "utils/lsyscache.h"
#include "utils/typcache.h"
#include "utils/builtins.h"
#include "utils/index_selfuncs.h"
#include "miscadmin.h"
#include "tcop/utility.h"
#include "postmaster/bgworker.h"
#include "pgstat.h"
#include "executor/executor.h"
#include "storage/ipc.h"
#include "storage/latch.h"
#include "storage/lock.h"
#include "storage/lmgr.h"
#include "storage/procarray.h"

#include "lsm.h"

#ifdef PG_MODULE_MAGIC
PG_MODULE_MAGIC;
#endif

PG_FUNCTION_INFO_V1(lsm_handler);


/* Initialize lsm control data entry */
static void lsm_init_entry(LsmMetaData* entry, Relation index)
{
	entry->n_inserts = 0;
	entry->l1= InvalidOid;
	entry->l0=index->rd_id;
	elog(NOTICE,"INIT writing l0 %d",entry->l0);
	entry->heap = index->rd_index->indrelid;
	entry->db_id = MyDatabaseId;
	entry->user_id = GetUserId();
	entry->top_index_size = 64;
}

/* Get B-Tree index size (number of blocks) */
static BlockNumber lsm_get_index_size(Oid relid)
{
       Relation index = index_open(relid, AccessShareLock);
       BlockNumber size = RelationGetNumberOfBlocks(index);
	   index_close(index, AccessShareLock);
       return size;
}

//			char* topidxname = psprintf("%s_top%d", relname, i);
/* Truncate top index */
static void
lsm_truncate_index(Relation index, Oid heap_oid)
{
	elog(NOTICE,"truncate heap rel Oid %d",heap_oid);
	//Relation index = index_open(index_oid, AccessExclusiveLock);
	Relation heap = table_open(heap_oid, AccessShareLock); /* heap is actually not used, because we will not load data to top indexes */
	IndexInfo* indexInfo = BuildDummyIndexInfo(index);
	RelationTruncate(index, 0);
	elog(LOG, "lsm: truncate index %s", RelationGetRelationName(index));
	/* parameter third is for reindex it called ambuild internally*/
	index_build(heap, index, indexInfo, true, false);
	/* rewrite meta*/ 
	//index_close(index, AccessExclusiveLock);
	table_close(heap, AccessShareLock);
}

/* Merge top index into base index */
static void
lsm_merge_indexes(Oid dst_oid, Relation top_index, Oid heap_oid)
{
	elog(NOTICE,"merge dest_oid %d,heap rel Oid %d",dst_oid,heap_oid);
	Relation heap = table_open(heap_oid, AccessShareLock);
	Relation base_index = index_open(dst_oid, RowExclusiveLock);
	IndexScanDesc scan;
	bool ok;
	Oid  save_am = base_index->rd_rel->relam;

	elog(LOG, "lsm: merge top index %s with size %d blocks", RelationGetRelationName(top_index), RelationGetNumberOfBlocks(top_index));

	base_index->rd_rel->relam = BTREE_AM_OID;
	scan = index_beginscan(heap, top_index, SnapshotAny, 0, 0);
	scan->xs_want_itup = true;
	btrescan(scan, NULL, 0, 0, 0);
	for (ok = _bt_first(scan, ForwardScanDirection); ok; ok = _bt_next(scan, ForwardScanDirection))
	{
		IndexTuple itup = scan->xs_itup;
		if (BTreeTupleIsPosting(itup))
		{
			/* 
			 * If index tuple is posting item, we need to transfer it to normal index tuple.
			 * Posting list is representing by index tuple with INDEX_ALT_TID_MASK bit set in t_info and
			 * BT_IS_POSTING bit in TID offset, following by array of TIDs.
			 * We need to store right TID (taken from xs_heaptid) and correct index tuple length
			 * (not including size of TIDs array), clearing INDEX_ALT_TID_MASK.
			 * For efficiency reasons let's do it in place, saving and restoring original values after insertion is done.
			 */
			ItemPointerData save_tid = itup->t_tid;
			unsigned short save_info = itup->t_info;
			itup->t_info = (save_info & ~(INDEX_SIZE_MASK | INDEX_ALT_TID_MASK)) + BTreeTupleGetPostingOffset(itup);
			itup->t_tid = scan->xs_heaptid;
			_bt_doinsert(base_index, itup, false, heap); /* lsm index is not unique so need not to heck for duplica
tes */
			itup->t_tid = save_tid;
			itup->t_info = save_info;
		}
		else
		{
			_bt_doinsert(base_index, itup, false, heap); /* lsm index is not unique so need not to heck for duplica
tes */
		}
	}
	index_endscan(scan);
	base_index->rd_rel->relam = save_am;
	index_close(base_index, RowExclusiveLock);
	table_close(heap, AccessShareLock);
}
/*
 * lsm access methods implementation
 */

void
lsmbuildempty( Relation index)
{
	/*
	This function is copied from nbtree.c and our metdata is added to existing metadata
	*/ 
	Page		metapage;
	BTMetaPageData* metad;
	LsmMetaData* lsmMeta;
	Oid save_am;
	save_am = index->rd_rel->relam;

	/* Construct metapage. */
	metapage = (Page) palloc(BLCKSZ);
	index->rd_rel->relam = BTREE_AM_OID;
	_bt_initmetapage(metapage, P_NONE, 0, _bt_allequalimage(index, false));
	index->rd_rel->relam = save_am;
	
	metad = BTPageGetMeta(metapage);
	lsmMeta=(LsmMetaData*)(metad+1);

	lsm_init_entry(lsmMeta,index);
	((PageHeader) metapage)->pd_lower =
		((char *) metad + sizeof(BTMetaPageData)) +sizeof(LsmMetaData) - (char *) metapage;
	/*
	 * Write the page and log it.  It might seem that an immediate sync would
	 * be sufficient to guarantee that the file exists on disk, but recovery
	 * itself might remove it while replaying, for example, an
	 * XLOG_DBASE_CREATE or XLOG_TBLSPC_CREATE record.  Therefore, we need
	 * this even when wal_level=minimal.
	 */
	PageSetChecksumInplace(metapage, BTREE_METAPAGE);
	smgrwrite(index->rd_smgr, INIT_FORKNUM, BTREE_METAPAGE,
			  (char *) metapage, true);
	log_newpage(&index->rd_smgr->smgr_rnode.node, INIT_FORKNUM,
				BTREE_METAPAGE, metapage, true);

	/*
	 * An immediate sync is required even if we xlog'd the page, because the
	 * write did not go through shared_buffers and therefore a concurrent
	 * checkpoint may have moved the redo pointer past our xlog record.
	 */
	smgrimmedsync(index->rd_smgr, INIT_FORKNUM);
}
IndexBuildResult *
lsmbuild(Relation heap, Relation index, IndexInfo *indexInfo)
{
	Buffer	metabuf;
	Page metapg;
	BTMetaPageData* metad;
	LsmMetaData* lsmMeta;
	IndexBuildResult* result;
	Oid save_am;
	save_am = index->rd_rel->relam;
	elog(NOTICE,"LSM Build");
	index->rd_rel->relam = BTREE_AM_OID;
	result=btbuild( heap, index, indexInfo);
	index->rd_rel->relam = save_am;
	elog(NOTICE,"build return now getting metadata");
	metabuf = _bt_getbuf(index, BTREE_METAPAGE, BT_WRITE);
	metapg = BufferGetPage(metabuf);
	metad = BTPageGetMeta(metapg);
	lsmMeta=(LsmMetaData*)(metad+1);
	lsm_init_entry(lsmMeta,index);
	elog(NOTICE,"LO created oid %d",lsmMeta->l0);
	lsmMeta->n_inserts=lsmMeta->n_inserts+result->index_tuples;
	elog(NOTICE,"No of tuple generated by btreebuild %ld ", lsmMeta->n_inserts);
	_bt_relbuf(index, metabuf);

	
		return result;
}
void
lsm_create_l1_if_not_exits(Relation heap,Relation index,LsmMetaData* lsmMetaCopy){
	char* newName;
	char* l0name;
	Relation l1;
	if(lsmMetaCopy->l1==InvalidOid){
			l0name=index->rd_rel->relname.data;
			newName=(char*)palloc(NAMEDATALEN+2);
			newName[0]='L';
			newName[1]='1';
		for(int i=0;i<NAMEDATALEN;++i){
			newName[i+2]=l0name[i];
		}
		lsmMetaCopy->l1= index_concurrently_create_copy(
			heap,/*heap relation*/
			index->rd_id,/*old oid*/
			newName/*new name =l1+oldname*/
		);
		l1=index_open(lsmMetaCopy->l1,AccessShareLock);
		index_build(heap,l1,BuildIndexInfo(l1),false,false);
		index_close(l1,AccessShareLock);
	}
}
/* Insert in active top index, on overflow take data from lo and initiate merge to base index  l1*/
static bool
lsm_insert(Relation rel, Datum *values, bool *isnull,
			ItemPointer ht_ctid, Relation heapRel,
			IndexUniqueCheck checkUnique,
			IndexInfo *indexInfo)
{
	bool result;
	Buffer	metabuf;
	Page metapg;
	BTMetaPageData* metad;
	LsmMetaData* lsmMeta;
	LsmMetaData* lsmMetaCopy;
	Oid save_am;
	save_am = rel->rd_rel->relam;

	elog(NOTICE,"lsm Insert calling btree insert");

	rel->rd_rel->relam = BTREE_AM_OID;
	result=btinsert( rel,values, isnull,  ht_ctid,  heapRel, checkUnique,indexInfo);
	rel->rd_rel->relam = save_am;

	metabuf = _bt_getbuf(rel, BTREE_METAPAGE, BT_WRITE);
	metapg = BufferGetPage(metabuf);
	
	metad = BTPageGetMeta(metapg);
	lsmMeta=(LsmMetaData*)(metad+1);
	elog(NOTICE,"BT META START %p BT META SIZE %ld LSM META START %p lo from meta %d prev size %ld",metad,sizeof(BTMetaPageData),lsmMeta,lsmMeta->l0,lsmMeta->n_inserts);
	lsmMeta->n_inserts=lsmMeta->n_inserts+1;
	elog(NOTICE,"No of tuple generated by btree build %ld ", lsmMeta->n_inserts);
	if(lsmMeta->n_inserts > 2){
		elog(NOTICE,"if part");
		lsmMetaCopy=(LsmMetaData*)palloc(sizeof(LsmMetaData));
		elog(NOTICE,"palloc success");
		memcpy(lsmMetaCopy,lsmMeta,sizeof(LsmMetaData));
		elog(NOTICE,"memcpy success ");
		_bt_relbuf(rel, metabuf);
		elog(NOTICE,"Max no of inserts 64+ achieved Closed relation");

	/*handling of Merging of top and bottom index*/
			//StartTransactionCommand();
			//{

				pgstat_report_activity(STATE_RUNNING, "merging");
				elog(NOTICE,"creation of li if not found");
				lsm_create_l1_if_not_exits(heapRel,rel, lsmMetaCopy);
				elog(NOTICE,"lo oid in meta %d  lo oid from rel %d l1 oid  %d  merging heap rel oid %d", lsmMetaCopy->l0,rel->rd_id,lsmMetaCopy->l1,lsmMetaCopy->heap);
				elog(NOTICE,"merging ");
				lsm_merge_indexes(lsmMetaCopy->l1, rel, lsmMetaCopy->heap);
				elog(NOTICE,"merging success");
				elog(NOTICE,"trucncate");
				pgstat_report_activity(STATE_RUNNING, "truncate And rebuild top index");

				lsm_truncate_index(rel,lsmMetaCopy->heap);
				elog(NOTICE,"maintaing new meta");
					metabuf = _bt_getbuf(rel, BTREE_METAPAGE, BT_WRITE);
					metapg = BufferGetPage(metabuf);
					metad = BTPageGetMeta(metapg);
					lsmMeta=(LsmMetaData*)(metad+1);
					lsmMeta->n_inserts=0;
					lsmMeta->top_index_size=lsmMetaCopy->top_index_size;
					lsmMeta->user_id=lsmMetaCopy->user_id;
					lsmMeta->db_id=lsmMetaCopy->db_id;
					lsmMeta->heap=lsmMetaCopy->heap;
					lsmMeta->l1=lsmMetaCopy->l1;
					lsmMeta->l0=lsmMetaCopy->l0;
					_bt_relbuf(rel, metabuf);
				pfree(lsmMetaCopy);
				elog(NOTICE,"insert and merging done");
			//}
			//CommitTransactionCommand();

	}else{
		elog(NOTICE,"Returing from insert in LO because Size not Exceed");
			_bt_relbuf(rel, metabuf);
	}
	return result;
}


Datum lsm_handler(PG_FUNCTION_ARGS)
{
	elog(NOTICE,"LSM HANDLER FUCNCTION CALL");
	pg_printf("LSM HANDLER PSPRINTF");
	IndexAmRoutine *amroutine;
	amroutine= makeNode(IndexAmRoutine);

	amroutine->amstrategies = BTMaxStrategyNumber;
	amroutine->amsupport = BTNProcs;
	amroutine->amoptsprocnum = BTOPTIONS_PROC;
	amroutine->amcanorder = true;
	amroutine->amcanorderbyop = true;
	amroutine->amcanbackward = true;
	amroutine->amcanunique = false;   /* We can't check that index is unique without accessing base index */
	amroutine->amcanmulticol = true;
	amroutine->amoptionalkey = true;
	amroutine->amsearcharray = false; /* TODO: not sure if it will work correctly with merge */
	amroutine->amsearchnulls = true;
	amroutine->amstorage = false;
	amroutine->amclusterable = true;
	amroutine->ampredlocks = true;
	amroutine->amcanparallel = false; /* TODO: parallel scac is not supported yet */
	amroutine->amcaninclude = true;
	amroutine->amusemaintenanceworkmem = false;
	amroutine->amparallelvacuumoptions = 	VACUUM_OPTION_PARALLEL_BULKDEL | VACUUM_OPTION_PARALLEL_COND_CLEANUP;;
	amroutine->amkeytype = InvalidOid;

	amroutine->ambuild = lsmbuild;
	amroutine->ambuildempty = lsmbuildempty;
	amroutine->aminsert = lsm_insert;
	amroutine->ambulkdelete = btbulkdelete;
	amroutine->amvacuumcleanup = btvacuumcleanup;
	amroutine->amcanreturn = btcanreturn;
	amroutine->amcostestimate = btcostestimate;
	amroutine->amoptions = btoptions;
	amroutine->amproperty = btproperty;
	amroutine->ambuildphasename = btbuildphasename;
	amroutine->amvalidate = btvalidate;
	amroutine->ambeginscan = btbeginscan;
	amroutine->amrescan = btrescan;
	amroutine->amgettuple = btgettuple;
	amroutine->amgetbitmap = btgetbitmap;
	amroutine->amendscan = btendscan;
	amroutine->ammarkpos = NULL;  /*  When do we need index_markpos? Can we live without it? */
	amroutine->amrestrpos = NULL;
	amroutine->amestimateparallelscan = NULL;
	amroutine->aminitparallelscan = NULL;
	amroutine->amparallelrescan = NULL;

	PG_RETURN_POINTER(amroutine);
}